#include "SkyPreetham.h"
#include "common.h"
#include "Vector3.h"
#include "SceneEntity.h"
#include "Transform3.h"
#include "Camera.h"
#include "RenderState.h"

using namespace CHCDemoEngine;
using namespace std;



static CGprogram sCgSkyProgram = NULL;
static CGparameter sLightDirParam;
static CGparameter sThetaSunParam;
static CGparameter sZenithColorParam;
static CGparameter sAColorParam;
static CGparameter sBColorParam;
static CGparameter sCColorParam;
static CGparameter sDColorParam;
static CGparameter sEColorParam;


inline float CBQ(float x)
{
	return x * x * x;
}


inline float SQR(float x)
{
	return x * x ;
}


void SkyPreetham::Init(CGcontext context)
{	
	sCgSkyProgram = 
		cgCreateProgramFromFile(context, 
								CG_SOURCE,
								"src/shaders/sky_preetham.cg", 
								RenderState::sCgVertexProfile,
								"default_vs",
								NULL);

	if (sCgSkyProgram != NULL)
	{
		cgGLLoadProgram(sCgSkyProgram);

		sLightDirParam = cgGetNamedParameter(sCgSkyProgram, "lightDir");
		sThetaSunParam = cgGetNamedParameter(sCgSkyProgram, "thetaSun");
		sZenithColorParam = cgGetNamedParameter(sCgSkyProgram, "zenithColor");
		sAColorParam = cgGetNamedParameter(sCgSkyProgram, "aColor");
		sBColorParam = cgGetNamedParameter(sCgSkyProgram, "bColor");
		sCColorParam = cgGetNamedParameter(sCgSkyProgram, "cColor");
		sDColorParam = cgGetNamedParameter(sCgSkyProgram, "dColor");
		sEColorParam = cgGetNamedParameter(sCgSkyProgram, "eColor");
	}
	else
		cerr << "sky program failed to load" << endl;
}


SkyPreetham::SkyPreetham(float turbitity, SceneEntity *skyDome):
mSkyDome(skyDome),
mTurbidity(turbitity)
//	mSunQuad(NULL)
{
	CreateSunQuad();
}


SkyPreetham::~SkyPreetham()
{
	//DEL_PTR(mSunQuad);
}


void SkyPreetham::Compute(const Vector3 &sunDir, Camera *camera, RenderState *state)
{
	float sun_theta[2];

	sun_theta[0] = acos(sunDir.z);

	const float cos_theta = cos(sun_theta[0]);
	sun_theta[1] = cos_theta * cos_theta;

	Vector3 zenithColor;

	zenithColor.x = ( 0.00165f * CBQ(sun_theta[0]) - 0.00374f * SQR(sun_theta[0]) + 0.00208f * sun_theta[0] + 0.0f)     * SQR(mTurbidity) +	
				    (-0.02902f * CBQ(sun_theta[0]) + 0.06377f * SQR(sun_theta[0]) - 0.03202f * sun_theta[0] + 0.00394f) * mTurbidity +
					( 0.11693f * CBQ(sun_theta[0]) - 0.21196f * SQR(sun_theta[0]) + 0.06052f * sun_theta[0] + 0.25885f);

	zenithColor.y = (0.00275f  * CBQ(sun_theta[0]) - 0.00610f * SQR(sun_theta[0]) + 0.00316f * sun_theta[0] + 0.0f)     * SQR(mTurbidity) +
	                (-0.04214f * CBQ(sun_theta[0]) + 0.08970f * SQR(sun_theta[0]) - 0.04153f * sun_theta[0] + 0.00515f) * mTurbidity +
		            ( 0.15346f * CBQ(sun_theta[0]) - 0.26756f * SQR(sun_theta[0]) + 0.06669f * sun_theta[0] + 0.26688f);
	
	zenithColor.z  = (float)((4.0453f * mTurbidity - 4.9710f) *	tan((4.0f / 9.0f - mTurbidity / 120.0f) * 
							 (M_PI - 2.0f * sun_theta[0])) - 0.2155f * mTurbidity + 2.4192f);

	// convert kcd/m to cd/m
	zenithColor.z *= 1000.0f;

	double ABCDE_x[5], ABCDE_y[5], ABCDE_Y[5];

	ABCDE_x[0] = -0.01925 * mTurbidity - 0.25922;
	ABCDE_x[1] = -0.06651 * mTurbidity + 0.00081;
	ABCDE_x[2] = -0.00041 * mTurbidity + 0.21247;
	ABCDE_x[3] = -0.06409 * mTurbidity - 0.89887;
	ABCDE_x[4] = -0.00325 * mTurbidity + 0.04517;

	ABCDE_y[0] = -0.01669 * mTurbidity - 0.26078;
	ABCDE_y[1] = -0.09495 * mTurbidity + 0.00921;
	ABCDE_y[2] = -0.00792 * mTurbidity + 0.21023;
	ABCDE_y[3] = -0.04405 * mTurbidity - 1.65369;
	ABCDE_y[4] = -0.01092 * mTurbidity + 0.05291;

	ABCDE_Y[0] =  0.17872 * mTurbidity - 1.46303;
	ABCDE_Y[1] = -0.35540 * mTurbidity + 0.42749;
	ABCDE_Y[2] = -0.02266 * mTurbidity + 5.32505;
	ABCDE_Y[3] =  0.12064 * mTurbidity - 2.57705;
	ABCDE_Y[4] = -0.06696 * mTurbidity + 0.37027;


	// Move skybox with camera.
	Vector3 position = camera->GetPosition();

	const float scaleFactor = 100.0f;

	position.z -= 10 * scaleFactor;
	Matrix4x4 m = TranslationMatrix(position);

	Matrix4x4 s = ScaleMatrix(scaleFactor, scaleFactor, scaleFactor);
	mSkyDome->GetTransform()->SetMatrix(s * m);
	
	cgGLSetParameter3f(sLightDirParam, sunDir.x, sunDir.y, sunDir.z);
	cgGLSetParameter2f(sThetaSunParam, sun_theta[0], sun_theta[1]);
	cgGLSetParameter3f(sZenithColorParam, zenithColor.x, zenithColor.y, zenithColor.z);

	cgGLSetParameter3f(sAColorParam, ABCDE_x[0], ABCDE_y[0], ABCDE_Y[0]);
	cgGLSetParameter3f(sBColorParam, ABCDE_x[1], ABCDE_y[1], ABCDE_Y[1]);
	cgGLSetParameter3f(sCColorParam, ABCDE_x[2], ABCDE_y[2], ABCDE_Y[2]);
	cgGLSetParameter3f(sDColorParam, ABCDE_x[3], ABCDE_y[3], ABCDE_Y[3]);
	cgGLSetParameter3f(sEColorParam, ABCDE_x[4], ABCDE_y[4], ABCDE_Y[4]);

	cgGLEnableProfile(RenderState::sCgVertexProfile);
	cgGLBindProgram(sCgSkyProgram);
	
	

	// Render sky dome.
	mSkyDome->Render(state);

	// Render additively blended sun disc.
	//RenderSunDisk(sunDir, camera);
}


void SkyPreetham::RenderSunDisk(const Vector3 &sunDir, Camera *camera)
{
	// Move skybox with camera.
	Vector3 position = camera->GetPosition();

	const float scaleFactor = 100.0f;
	position.z -= 10 * scaleFactor;

	// Set world matrix to sun disc position.
	Matrix4x4 sunPos = TranslationMatrix(position);

	Vector3 ndir = -Normalize(sunDir);

	const float pitch = -atan2(ndir.x, ndir.y);
	const float yaw = atan2(ndir.z, sqrt((ndir.x * ndir.x) + (ndir.y * ndir.y)));

	Matrix4x4 roty = RotationYMatrix(pitch);
	Matrix4x4 rotx = RotationXMatrix(yaw);

	sunPos *= roty;
	sunPos *= rotx;;

	//sunposition.rotateAroundY(-D().getScene().getLight().getHorizontalOrientation() + 90.0f);
	//sunposition.rotateAroundX(-D().getScene().getLight().getVerticalOrientation());
	
	glMatrixMode(GL_MODELVIEW);
	glMultMatrixf((float *)sunPos.x);

	float ambient[] = {1.0f, 1.0f, 1.0f, 1.0f};

	glMaterialfv(GL_FRONT, GL_AMBIENT, ambient);

	//Float size = 0.5f + (1.0f - lightdirection.y) * 0.5f;

	//mSunEffect->setVariableVector3("LightDiffuseColor", D().getScene().getLight().getDiffuseColor());
	//mSunEffect->setVariableFloat("SunSize", size);
	//mSunEffect->setVariableTexture2D("SunTexture", mSunTexture);

	/*mSunEffect->activate();
	
	mSunQuad->render();

	mSunEffect->deactivate();
	F().getRenderDevice().setDefaultBlendingMode(FRenderDevice::BLENDING_NONE);*/

	glPopMatrix();
}


void SkyPreetham::CreateSunQuad()
{
	/*
	mSunQuad = new FVertexBuffer();
	mSunQuad->setupPrimitiveType(FVertexBuffer::PRIMITIVES_TRIANGLES);
	mSunQuad->setupVertexFormat(3, 0, 0, true);
	mSunQuad->setupTexCoordSet(0, 2);
	mSunQuad->setVertexBufferSize(4, 6);

	mSunQuad->setVertexPosition(0, FVector3(-0.1f,  0.1f, 1.0f));
	mSunQuad->setVertexPosition(1, FVector3( 0.1f,  0.1f, 1.0f));
	mSunQuad->setVertexPosition(2, FVector3(-0.1f, -0.1f, 1.0f));
	mSunQuad->setVertexPosition(3, FVector3( 0.1f, -0.1f, 1.0f));
	mSunQuad->setVertexTexCoord(0, 0, FVector2(0.0f, 0.0f));
	mSunQuad->setVertexTexCoord(1, 0, FVector2(1.0f, 0.0f));
	mSunQuad->setVertexTexCoord(2, 0, FVector2(0.0f, 1.0f));
	mSunQuad->setVertexTexCoord(3, 0, FVector2(1.0f, 1.0f));

	mSunQuad->setIndex(0, 0);
	mSunQuad->setIndex(1, 1);
	mSunQuad->setIndex(2, 2);
	mSunQuad->setIndex(3, 2);
	mSunQuad->setIndex(4, 1);
	mSunQuad->setIndex(5, 3);
	*/
}